/*
 * Created on 04.01.2006
 *
 * Copyright --- All rights reserved.
 * ITI, Universität Karlsruhe, Germany.
 */
package de.uka.algo.GT3D;

import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import org.xml.sax.Attributes;
import org.xml.sax.SAXException;

import de.uka.algo.GT3D.exceptions.FileFormatException;
import edu.uci.ics.jung.graph.Edge;
import edu.uci.ics.jung.graph.Graph;
import edu.uci.ics.jung.graph.Vertex;
import edu.uci.ics.jung.graph.decorators.StringLabeller;
import edu.uci.ics.jung.graph.impl.DirectedSparseGraph;
import edu.uci.ics.jung.graph.impl.SparseVertex;
import edu.uci.ics.jung.graph.impl.UndirectedSparseGraph;
import edu.uci.ics.jung.io.GraphMLFileHandler;
import edu.uci.ics.jung.utils.GraphUtils;
import edu.uci.ics.jung.utils.UserData;

/**
 * A GraphML file handler to parse simple GraphML files
 * 
 * @author stm
 *
 */
public class GT3DGraphMLFileHandler extends GraphMLFileHandler {

	
	/**
	 * The graph we are reading.
	 */
	private Graph mGraph;
    /**
	 * Stores the node and edge labels that can be retrieved by its
	 * <code>getUserData</code> method.
	 */
    private StringLabeller mLabeller;

    
    /**
     * The edge represented by the currently open edge tag or null.
     */
    Edge currentEdge= null;
    /**
     * The node represented by the currently open edge tag or null.
     */
    Vertex currentVertex = null;
    /**
     * The node represented by the currently open data tag or null.
     */
    String currentKey = null;
    
    
    /**
     * A map from keys (of data tags) to the corresponding names.
     */
    Map<String, String> keyMap = null;
    
    public GT3DGraphMLFileHandler() {
		super();
	}

    protected Graph getGraph() {
        return mGraph;
    }

    protected StringLabeller getLabeller() {
        return mLabeller;
    }

    private Map<String, String> getAttributeMap(Attributes attrs) {
        Map<String, String> map = new HashMap<String, String>();
        if (attrs != null) {
            for (int i = 0; i < attrs.getLength(); i++) {
                map.put(attrs.getQName(i), attrs.getValue(i));
            }
        }
        return map;
    }

    
    /**
     * creates a new graph
     * 
     * @param attributeMap the attributes of the <tt>graph</tt> tag.
     * @see edu.uci.ics.jung.io.GraphMLFileHandler#createGraph(java.util.Map)
     */
    protected void createGraph(Map attributeMap) {
        String edgeDefaultType =
                (String) attributeMap.remove("edgedefault");
        if (edgeDefaultType.equals("directed")) {
            mGraph = new DirectedSparseGraph();
        } else if (edgeDefaultType.equals("undirected")) {
            mGraph = new UndirectedSparseGraph();
        } else {
            throw new FileFormatException("Error parsing graph. Edge default type not specified.");
        }

        mLabeller = StringLabeller.getLabeller(mGraph);

        for (Iterator keyIt = attributeMap.keySet().iterator(); keyIt.hasNext();) {
            Object key = keyIt.next();
            Object value = attributeMap.get(key);
            mGraph.setUserDatum(key, value, UserData.SHARED);
        }

    }

    /**
	 * Initialize a new vertex and insert it into the graph. This method 
	 * is called if a <tt>node</tt> tag is opened. A node is identified by
	 * its <tt>id</tt> tag and may contain <tt>data</tt> tags.
	 * 
	 * @param attributeMap the attributes of the <tt>node</tt> tag.
	 * @return The newly created vertex.
	 */
    protected Vertex startVertex(Map attributeMap) {
        if (mGraph == null) {
            throw new FileFormatException("Error parsing graph. Graph element must be specified before node element.");
        }

        // add a new vertex into the graph
        Vertex vertex = mGraph.addVertex(new SparseVertex());
        String idString = (String) attributeMap.remove("id");

        try {
            mLabeller.setLabel((Vertex) vertex,idString);
        } catch (StringLabeller.UniqueLabelException ule) {
            throw new FileFormatException("Ids must be unique");

        }

        return vertex;
    }

    /**
	 * Initialize a new edge and insert it into the graph. This method 
	 * is called if an <tt>edge</tt> tag is opened. An edge is identified by
	 * its <tt>id</tt> tag and may contain <tt>data</tt> tags.
	 * 
	 * @param attributeMap the attributes of the <tt>edge</tt> tag.
	 * @return the newly created edge.
	 */
    protected Edge startEdge(Map attributeMap) {
        if (mGraph == null) {
            throw new FileFormatException("Error parsing graph. Graph element must be specified before edge element.");
        }

        // get the source and target attributes
        String sourceId = (String) attributeMap.remove("source");
        Vertex sourceVertex =
                mLabeller.getVertex(sourceId);

        String targetId = (String) attributeMap.remove("target");
        Vertex targetVertex =
                 mLabeller.getVertex(targetId);

        // insert a new edge
        Edge e = null;
        if (targetVertex != null && sourceVertex != null)
        	e = GraphUtils.addEdge(mGraph, sourceVertex, targetVertex);
    	if (e == null)
    		throw new FileFormatException("Error parsing graph. GraphML file invalid.");


        return e;
    }

    
    /**
     * Called when a <tt>vertex</tt> tag is closed.
     */
    protected void endVertex(){
    	if (currentVertex == null)
    		throw new FileFormatException("Error parsing graph. GraphML file invalid.");
    		
    	currentVertex = null;
    }
    
    /**
     * Called when a <tt>edge</tt> tag is closed.
     */
    protected void endEdge(){
    	if (currentEdge == null)
    		throw new FileFormatException("Error parsing graph. GraphML file invalid.");
    		
    	currentEdge = null;
    }
    
    /**
     * Inserts a new (key, value)-pair into <data>keyMap</data>. Called when
     * a <tt>key</tt> tag is opened. A <tt>key</tt> tag has the attributes
     * <tt>id</tt> and <tt>for</tt> and additional <tt>attr.type</tt> and 
     * <tt>attr.name</tt> attributes. 
     * 
     * @param attributeMap The attributes of the <tt>key</tt> tag
     */
    protected void insertKey(Map<String, String> attributeMap) {
    
    	String id = (String) attributeMap.get("id");
    	String target = (String) attributeMap.get("for");
    	String type = (String) attributeMap.get("attr.type");
    	String name = (String) attributeMap.get("attr.name");
    	
    	keyMap.put(id, name);
    	
    }
    
    /**
     * Called if a <tt>data</tt> tag is found. This tag is identified by a
     * <tt>key</tt> attribute which is specified in a <tt>key</tt> entry
     * and contains the edge or node label as content.
     * 
     * @param attributeMap The attributes of the tag.
     */
    protected void startData(Map attributeMap) {
    	String key = (String) attributeMap.get("key");
    	
    	// lookup the key in keyMap
    	if (currentEdge!= null || currentVertex != null)
    		currentKey = keyMap.get(key);
    }
    
    /**
     * Called if a <tt>data</tt> tag is closed.
     */
    protected void endData() {
    	currentKey = null;
    }
    
    /**
     * This method is called whenever the SaX parser encounters an opening 
     * tag
     * 
     * @see org.xml.sax.ContentHandler#startElement(java.lang.String, java.lang.String, java.lang.String, org.xml.sax.Attributes)
     */
    public void startElement(
            String lName,
            String namespaceURI,
            // local name
            String qName, // qualified name
            Attributes attrs) throws SAXException {

    	// parse attributes
        Map<String, String> attributeMap = getAttributeMap(attrs);

        if (qName.toLowerCase().equals("graphml")) {
            // the <graphml> tag is the root tag of the document
            // --> initialize data structures
        	keyMap = new HashMap<String, String>();
        	
        }else if (qName.toLowerCase().equals("graph")) {
        	// the <graph> tag encloses the <node> and <edge> tags
            createGraph(attributeMap);
            
        } else if (qName.toLowerCase().equals("node")) {
        	// a node tag 
        	if (currentEdge != null || currentVertex != null)
        		throw new FileFormatException("Error parsing graph. GraphML file invalid.");
            currentVertex=startVertex(attributeMap);

        } else if (qName.toLowerCase().equals("edge")) {
        	// an edge tag
        	if (currentEdge != null || currentVertex != null)
        		throw new FileFormatException("Error parsing graph. GraphML file invalid.");
            currentEdge=startEdge(attributeMap);

        } else if (qName.toLowerCase().equals("key")) {
        	// a key tag
        	insertKey(attributeMap);
        
        } else if (qName.toLowerCase().equals("data")) {
        	// a data tag
        	startData(attributeMap);
        }
    }

    
    /**
     * This method is called whenever the SaX parser encounters a closing 
     * tag
     * 
     * @see org.xml.sax.ContentHandler#startElement(java.lang.String, java.lang.String, java.lang.String, org.xml.sax.Attributes)
     */
    public void endElement(String uri,
            String localName,
            String qName){
        if (qName.toLowerCase().equals("graph")) {
        	// do nothing
        } else if (qName.toLowerCase().equals("node")) {
            endVertex();

        } else if (qName.toLowerCase().equals("edge")) {
            endEdge();
        } else if (qName.toLowerCase().equals("data")) {
            endData();
        }
    	
    }

    /**
     * This method is called whenever the SaX parser encounters 
     * character content inside a tag
     * 
     * @see org.xml.sax.ContentHandler#characters(char[], int, int)
     */
    public void characters(char[] ch,
            int start,
            int length){
    	
    	String content = new String(ch, start, length);
    	
    	if (currentKey == null){
    		return;
    	}
    	
    	if (currentVertex!=null){
    		currentVertex.setUserDatum(currentKey, content.trim(), UserData.SHARED);
    		currentKey = null;
    	} else 	if (currentEdge!=null){
    		currentEdge.setUserDatum(currentKey, content.trim(), UserData.SHARED);
    		currentKey = null;
    	} else
       		throw new FileFormatException("Error parsing graph. GraphML file invalid.");
    	
    }
    
}
